# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.time cimport time_t
from posix.types cimport mode_t, suseconds_t


cdef extern from "libssh/libssh.h" nogil:
    ctypedef struct fd_set:
        pass
    cdef struct timeval:
        time_t       tv_sec
        suseconds_t  tv_usec
    ctypedef unsigned char uint8_t
    ctypedef unsigned short uint16_t
    ctypedef unsigned int uint32_t
    ctypedef unsigned long long uint64_t
    ctypedef int socket_t
    int SSH_VERSION_INT(int, int, int)
    int SSH_VERSION_DOT(int, int, int)
    int SSH_VERSION(int, int, int)
    enum:
        LIBSSH_VERSION_MAJOR
        LIBSSH_VERSION_MINOR
        LIBSSH_VERSION_MICRO
    LIBSSH_VERSION_INT(int, int, int)
    LIBSSH_VERSION(int, int, int)
    struct ssh_counter_struct:
        uint64_t in_bytes
        uint64_t out_bytes
        uint64_t in_packets
        uint64_t out_packets
    # Forward declarations for structures
    # defined in other header files
    struct ssh_agent_struct:
        pass
    struct ssh_buffer_struct:
        pass
    struct ssh_channel_struct:
        pass
    struct ssh_message_struct:
        pass
    struct ssh_pcap_file_struct:
        pass
    struct ssh_key_struct:
        pass
    struct ssh_scp_struct:
        pass
    struct ssh_session_struct:
        pass
    struct ssh_string_struct:
        pass
    struct ssh_event_struct:
        pass
    struct ssh_connector_struct:
        pass
    ctypedef ssh_counter_struct *ssh_counter
    ctypedef ssh_agent_struct* ssh_agent
    ctypedef ssh_buffer_struct* ssh_buffer
    ctypedef ssh_channel_struct* ssh_channel
    ctypedef ssh_message_struct* ssh_message
    ctypedef ssh_pcap_file_struct* ssh_pcap_file
    ctypedef ssh_key_struct* ssh_key
    ctypedef ssh_scp_struct* ssh_scp
    ctypedef ssh_session_struct* ssh_session
    ctypedef ssh_string_struct* ssh_string
    ctypedef ssh_event_struct* ssh_event
    ctypedef ssh_connector_struct * ssh_connector
    ctypedef void* ssh_gssapi_creds
    enum:
        SSH_INVALID_SOCKET
    enum ssh_kex_types_e:
        SSH_KEX
        SSH_HOSTKEYS
        SSH_CRYPT_C_S
        SSH_CRYPT_S_C
        SSH_MAC_C_S
        SSH_MAC_S_C
        SSH_COMP_C_S
        SSH_COMP_S_C
        SSH_LANG_C_S
        SSH_LANG_S_C
    enum:
        SSH_CRYPT
        SSH_MAC
        SSH_COMP
        SSH_LANG
    enum ssh_auth_e:
        SSH_AUTH_SUCCESS,
        SSH_AUTH_DENIED,
        SSH_AUTH_PARTIAL,
        SSH_AUTH_INFO,
        SSH_AUTH_AGAIN,
        SSH_AUTH_ERROR
    # Auth flags
    enum:
        SSH_AUTH_METHOD_UNKNOWN
        SSH_AUTH_METHOD_NONE
        SSH_AUTH_METHOD_PASSWORD
        SSH_AUTH_METHOD_PUBLICKEY
        SSH_AUTH_METHOD_HOSTBASED
        SSH_AUTH_METHOD_INTERACTIVE
        SSH_AUTH_METHOD_GSSAPI_MIC
    # Messages
    enum ssh_requests_e:
        SSH_REQUEST_AUTH,
        SSH_REQUEST_CHANNEL_OPEN,
        SSH_REQUEST_CHANNEL,
        SSH_REQUEST_SERVICE,
        SSH_REQUEST_GLOBAL
    enum ssh_channel_type_e:
        SSH_CHANNEL_UNKNOWN,
        SSH_CHANNEL_SESSION,
        SSH_CHANNEL_DIRECT_TCPIP,
        SSH_CHANNEL_FORWARDED_TCPIP,
        SSH_CHANNEL_X11,
        SSH_CHANNEL_AUTH_AGENT
    enum ssh_channel_requests_e:
        SSH_CHANNEL_REQUEST_UNKNOWN,
        SSH_CHANNEL_REQUEST_PTY,
        SSH_CHANNEL_REQUEST_EXEC,
        SSH_CHANNEL_REQUEST_SHELL,
        SSH_CHANNEL_REQUEST_ENV,
        SSH_CHANNEL_REQUEST_SUBSYSTEM,
        SSH_CHANNEL_REQUEST_WINDOW_CHANGE,
        SSH_CHANNEL_REQUEST_X11
    enum ssh_global_requests_e:
        SSH_GLOBAL_REQUEST_UNKNOWN,
        SSH_GLOBAL_REQUEST_TCPIP_FORWARD,
        SSH_GLOBAL_REQUEST_CANCEL_TCPIP_FORWARD,
        SSH_GLOBAL_REQUEST_KEEPALIVE
    enum ssh_publickey_state_e:
        SSH_PUBLICKEY_STATE_ERROR,
        SSH_PUBLICKEY_STATE_NONE,
        SSH_PUBLICKEY_STATE_VALID,
        SSH_PUBLICKEY_STATE_WRONG
    enum:
        SSH_CLOSED
        SSH_READ_PENDING
        SSH_CLOSED_ERROR
        SSH_WRITE_PENDING
    enum ssh_server_known_e:
        SSH_SERVER_ERROR,
        SSH_SERVER_NOT_KNOWN,
        SSH_SERVER_KNOWN_OK,
        SSH_SERVER_KNOWN_CHANGED,
        SSH_SERVER_FOUND_OTHER,
        SSH_SERVER_FILE_NOT_FOUND
    enum:
        MD5_DIGEST_LEN
    # Errors
    enum ssh_error_types_e:
        SSH_NO_ERROR,
        SSH_REQUEST_DENIED,
        SSH_FATAL,
        SSH_EINTR
    enum ssh_keytypes_e:
        SSH_KEYTYPE_UNKNOWN,
        SSH_KEYTYPE_DSS,
        SSH_KEYTYPE_RSA,
        SSH_KEYTYPE_RSA1,
        SSH_KEYTYPE_ECDSA,  # Deprecated
        SSH_KEYTYPE_ED25519,
        SSH_KEYTYPE_DSS_CERT01,
        SSH_KEYTYPE_RSA_CERT01,
        SSH_KEYTYPE_ECDSA_P256,
        SSH_KEYTYPE_ECDSA_P384,
        SSH_KEYTYPE_ECDSA_P521,
        SSH_KEYTYPE_ECDSA_P256_CERT01,
        SSH_KEYTYPE_ECDSA_P384_CERT01,
        SSH_KEYTYPE_ECDSA_P521_CERT01,
        SSH_KEYTYPE_ED25519_CERT01
    enum ssh_keycmp_e:
        SSH_KEY_CMP_PUBLIC,
        SSH_KEY_CMP_PRIVATE
    enum:
        SSH_OK
        SSH_ERROR
        SSH_AGAIN
        SSH_EOF
    enum:
        SSH_LOG_NOLOG
        SSH_LOG_PROTOCOL
        SSH_LOG_PACKET
        SSH_LOG_FUNCTIONS
        SSH_LOG_RARE
    enum:
        SSH_LOG_NONE
        SSH_LOG_WARN
        SSH_LOG_INFO
        SSH_LOG_DEBUG
        SSH_LOG_TRACE
    enum ssh_options_e:
        SSH_OPTIONS_HOST,
        SSH_OPTIONS_PORT,
        SSH_OPTIONS_PORT_STR,
        SSH_OPTIONS_FD,
        SSH_OPTIONS_USER,
        SSH_OPTIONS_SSH_DIR,
        SSH_OPTIONS_IDENTITY,
        SSH_OPTIONS_ADD_IDENTITY,
        SSH_OPTIONS_KNOWNHOSTS,
        SSH_OPTIONS_TIMEOUT,
        SSH_OPTIONS_TIMEOUT_USEC,
        SSH_OPTIONS_SSH1,
        SSH_OPTIONS_SSH2,
        SSH_OPTIONS_LOG_VERBOSITY,
        SSH_OPTIONS_LOG_VERBOSITY_STR,
        SSH_OPTIONS_CIPHERS_C_S,
        SSH_OPTIONS_CIPHERS_S_C,
        SSH_OPTIONS_COMPRESSION_C_S,
        SSH_OPTIONS_COMPRESSION_S_C,
        SSH_OPTIONS_PROXYCOMMAND,
        SSH_OPTIONS_BINDADDR,
        SSH_OPTIONS_STRICTHOSTKEYCHECK,
        SSH_OPTIONS_COMPRESSION,
        SSH_OPTIONS_COMPRESSION_LEVEL,
        SSH_OPTIONS_KEY_EXCHANGE,
        SSH_OPTIONS_HOSTKEYS,
        SSH_OPTIONS_GSSAPI_SERVER_IDENTITY,
        SSH_OPTIONS_GSSAPI_CLIENT_IDENTITY,
        SSH_OPTIONS_GSSAPI_DELEGATE_CREDENTIALS,
        SSH_OPTIONS_HMAC_C_S,
        SSH_OPTIONS_HMAC_S_C,
        SSH_OPTIONS_PASSWORD_AUTH,
        SSH_OPTIONS_PUBKEY_AUTH,
        SSH_OPTIONS_KBDINT_AUTH,
        SSH_OPTIONS_GSSAPI_AUTH,
        SSH_OPTIONS_GLOBAL_KNOWNHOSTS,
        SSH_OPTIONS_NODELAY
    enum:
        SSH_SCP_WRITE
        SSH_SCP_READ
        SSH_SCP_RECURSIVE
    enum ssh_scp_request_types:
        SSH_SCP_REQUEST_NEWDIR,
        SSH_SCP_REQUEST_NEWFILE,
        SSH_SCP_REQUEST_EOF,
        SSH_SCP_REQUEST_ENDDIR,
        SSH_SCP_REQUEST_WARNING
    enum ssh_connector_flags_e:
        SSH_CONNECTOR_STDOUT,
        SSH_CONNECTOR_STDERR,
        SSH_CONNECTOR_BOTH
    int ssh_blocking_flush(ssh_session session, int timeout)
    ssh_channel ssh_channel_accept_x11(ssh_channel channel, int timeout_ms)
    int ssh_channel_change_pty_size(ssh_channel channel, int cols, int rows)
    int ssh_channel_close(ssh_channel channel)
    void ssh_channel_free(ssh_channel channel)
    int ssh_channel_get_exit_status(ssh_channel channel)
    ssh_session ssh_channel_get_session(ssh_channel channel)
    int ssh_channel_is_closed(ssh_channel channel)
    int ssh_channel_is_eof(ssh_channel channel)
    int ssh_channel_is_open(ssh_channel channel)
    ssh_channel ssh_channel_new(ssh_session session)
    int ssh_channel_open_auth_agent(ssh_channel channel)
    int ssh_channel_open_forward(ssh_channel channel, const char *remotehost, int remoteport, const char *sourcehost, int localport)
    int ssh_channel_open_session(ssh_channel channel)
    int ssh_channel_open_x11(ssh_channel channel, const char *orig_addr, int orig_port)
    int ssh_channel_poll(ssh_channel channel, int is_stderr)
    int ssh_channel_poll_timeout(ssh_channel channel, int timeout, int is_stderr)
    int ssh_channel_read(ssh_channel channel, void *dest, uint32_t count, int is_stderr)
    int ssh_channel_read_timeout(ssh_channel channel, void *dest,
                                 uint32_t count, int is_stderr, int timeout_ms)
    int ssh_channel_read_nonblocking(ssh_channel channel, void *dest, uint32_t count, int is_stderr)
    int ssh_channel_request_env(ssh_channel channel, const char *name, const char *value)
    int ssh_channel_request_exec(ssh_channel channel, const char *cmd)
    int ssh_channel_request_pty(ssh_channel channel)
    int ssh_channel_request_pty_size(ssh_channel channel, const char *term, int cols, int rows)
    int ssh_channel_request_shell(ssh_channel channel)
    int ssh_channel_request_send_signal(ssh_channel channel, const char *signum)
    int ssh_channel_request_send_break(ssh_channel channel, uint32_t length)
    int ssh_channel_request_sftp(ssh_channel channel)
    int ssh_channel_request_subsystem(ssh_channel channel, const char *subsystem)
    int ssh_channel_request_x11(ssh_channel channel, int single_connection, const char *protocol,
        const char *cookie, int screen_number)
    int ssh_channel_request_auth_agent(ssh_channel channel)
    int ssh_channel_send_eof(ssh_channel channel)
    int ssh_channel_select(ssh_channel *readchans, ssh_channel *writechans, ssh_channel *exceptchans, timeval * timeout)
    void ssh_channel_set_blocking(ssh_channel channel, int blocking)
    void ssh_channel_set_counter(ssh_channel channel, ssh_counter counter)
    int ssh_channel_write(ssh_channel channel, const void *data, uint32_t len)
    int ssh_channel_write_stderr(ssh_channel channel, const void *data, uint32_t len)
    uint32_t ssh_channel_window_size(ssh_channel channel)

    char *ssh_basename(const char *path)
    void ssh_clean_pubkey_hash(unsigned char **hash)
    int ssh_connect(ssh_session session)

    ssh_connector ssh_connector_new(ssh_session session)
    void ssh_connector_free(ssh_connector connector)
    int ssh_connector_set_in_channel(ssh_connector connector, ssh_channel channel, ssh_connector_flags_e flags)
    int ssh_connector_set_out_channel(ssh_connector connector, ssh_channel channel, ssh_connector_flags_e flags)
    void ssh_connector_set_in_fd(ssh_connector connector, socket_t fd)
    void ssh_connector_set_out_fd(ssh_connector connector, socket_t fd)

    const char *ssh_copyright()
    void ssh_disconnect(ssh_session session)
    char *ssh_dirname(const char *path)
    int ssh_finalize()

    # Reverse port forwarding
    ssh_channel ssh_channel_accept_forward(ssh_session session, int timeout_ms, int *destination_port)
    int ssh_channel_cancel_forward(ssh_session session, const char *address, int port)
    int ssh_channel_listen_forward(ssh_session session, const char *address, int port, int *bound_port)

    void ssh_free(ssh_session session)
    const char *ssh_get_disconnect_message(ssh_session session)
    const char *ssh_get_error(void *error)
    int ssh_get_error_code(void *error)
    socket_t ssh_get_fd(ssh_session session)
    char *ssh_get_hexa(const unsigned char *what, size_t len)
    char *ssh_get_issue_banner(ssh_session session)
    int ssh_get_openssh_version(ssh_session session)

    int ssh_get_server_publickey(ssh_session session, ssh_key *key)

    enum ssh_publickey_hash_type:
        SSH_PUBLICKEY_HASH_SHA1
        SSH_PUBLICKEY_HASH_MD5
    int ssh_get_publickey_hash(const ssh_key key, ssh_publickey_hash_type type, unsigned char **hash, size_t *hlen)

    # deprecated functions
    int ssh_get_pubkey_hash(ssh_session session, unsigned char **hash)
    ssh_channel ssh_forward_accept(ssh_session session, int timeout_ms)
    int ssh_forward_cancel(ssh_session session, const char *address, int port)
    int ssh_forward_listen(ssh_session session, const char *address, int port, int *bound_port)
    int ssh_get_publickey(ssh_session session, ssh_key *key)
    # End deprecated

    int ssh_get_random(void *where, int len, int strong)
    int ssh_get_version(ssh_session session)
    int ssh_get_status(ssh_session session)
    int ssh_get_poll_flags(ssh_session session)
    int ssh_init()
    int ssh_is_blocking(ssh_session session)
    int ssh_is_connected(ssh_session session)
    int ssh_is_server_known(ssh_session session)

    int ssh_set_log_level(int level)
    int ssh_get_log_level()
    void *ssh_get_log_userdata()
    int ssh_set_log_userdata(void *data)
    _ssh_log(int verbosity, const char *function, const char *format, int, int)

    # legacy
    ssh_log(ssh_session session, int prioriry, const char *format, int, int)

    ssh_channel ssh_message_channel_request_open_reply_accept(ssh_message msg)
    int ssh_message_channel_request_reply_success(ssh_message msg)
    void ssh_message_free(ssh_message msg)
    ssh_message ssh_message_get(ssh_session session)
    int ssh_message_subtype(ssh_message msg)
    int ssh_message_type(ssh_message msg)
    int ssh_mkdir(const char *pathname, mode_t mode)
    ssh_session ssh_new()

    int ssh_options_copy(ssh_session src, ssh_session *dest)
    int ssh_options_getopt(ssh_session session, int *argcptr, char **argv)
    int ssh_options_parse_config(ssh_session session, const char *filename)
    int ssh_options_set(ssh_session session, ssh_options_e type, const void *value)
    int ssh_options_get(ssh_session session, ssh_options_e type, char **value)
    int ssh_options_get_port(ssh_session session, unsigned int * port_target)
    int ssh_pcap_file_close(ssh_pcap_file pcap)
    void ssh_pcap_file_free(ssh_pcap_file pcap)
    ssh_pcap_file ssh_pcap_file_new()
    int ssh_pcap_file_open(ssh_pcap_file pcap, const char *filename)

    ctypedef int(*ssh_auth_callback)(const char *prompt, char *buf, size_t len, int echo, int verify, void *userdata)
    ssh_key ssh_key_new()
    void ssh_key_free(ssh_key key)
    ssh_keytypes_e ssh_key_type(const ssh_key key)
    const char *ssh_key_type_to_char(ssh_keytypes_e type)
    ssh_keytypes_e ssh_key_type_from_name(const char *name)
    int ssh_key_is_public(const ssh_key k)
    int ssh_key_is_private(const ssh_key k)
    int ssh_key_cmp(const ssh_key k1, const ssh_key k2, ssh_keycmp_e what)

    int ssh_pki_generate(ssh_keytypes_e type, int parameter, ssh_key *pkey)
    int ssh_pki_import_privkey_base64(const char *b64_key, const char *passphrase, ssh_auth_callback auth_fn, void *auth_data, ssh_key *pkey)
    int ssh_pki_import_privkey_file(const char *filename, const char *passphrase, ssh_auth_callback auth_fn, void *auth_data, ssh_key *pkey)
    int ssh_pki_export_privkey_file(const ssh_key privkey, const char *passphrase, ssh_auth_callback auth_fn, void *auth_data, const char *filename)

    int ssh_pki_copy_cert_to_privkey(const ssh_key cert_key, ssh_key privkey)

    int ssh_pki_import_pubkey_base64(const char *b64_key, ssh_keytypes_e type, ssh_key *pkey)
    int ssh_pki_import_pubkey_file(const char *filename, ssh_key *pkey)

    int ssh_pki_import_cert_base64(const char *b64_cert, ssh_keytypes_e type, ssh_key *pkey)
    int ssh_pki_import_cert_file(const char *filename, ssh_key *pkey)

    int ssh_pki_export_privkey_to_pubkey(const ssh_key privkey, ssh_key *pkey)
    int ssh_pki_export_pubkey_base64(const ssh_key key, char **b64_key)
    int ssh_pki_export_pubkey_file(const ssh_key key, const char *filename)

    const char *ssh_pki_key_ecdsa_name(const ssh_key key)

    char *ssh_get_fingerprint_hash(ssh_publickey_hash_type type, unsigned char *hash, size_t len)
    void ssh_print_hash(ssh_publickey_hash_type type, unsigned char *hash, size_t len)
    void ssh_print_hexa(const char *descr, const unsigned char *what, size_t len)
    int ssh_send_ignore(ssh_session session, const char *data)
    int ssh_send_debug(ssh_session session, const char *message, int always_display)
    void ssh_gssapi_set_creds(ssh_session session, const ssh_gssapi_creds creds)
    int ssh_scp_accept_request(ssh_scp scp)
    int ssh_scp_close(ssh_scp scp)
    int ssh_scp_deny_request(ssh_scp scp, const char *reason)
    void ssh_scp_free(ssh_scp scp)
    int ssh_scp_init(ssh_scp scp)
    int ssh_scp_leave_directory(ssh_scp scp)
    ssh_scp ssh_scp_new(ssh_session session, int mode, const char *location)
    int ssh_scp_pull_request(ssh_scp scp)
    int ssh_scp_push_directory(ssh_scp scp, const char *dirname, int mode)
    int ssh_scp_push_file(ssh_scp scp, const char *filename, size_t size, int perms)
    int ssh_scp_push_file64(ssh_scp scp, const char *filename, uint64_t size, int perms)
    int ssh_scp_read(ssh_scp scp, void *buffer, size_t size)
    const char *ssh_scp_request_get_filename(ssh_scp scp)
    int ssh_scp_request_get_permissions(ssh_scp scp)
    size_t ssh_scp_request_get_size(ssh_scp scp)
    uint64_t ssh_scp_request_get_size64(ssh_scp scp)
    const char *ssh_scp_request_get_warning(ssh_scp scp)
    int ssh_scp_write(ssh_scp scp, const void *buffer, size_t len)
    int ssh_select(ssh_channel *channels, ssh_channel *outchannels, socket_t maxfd, fd_set *readfds, timeval *timeout)
    int ssh_service_request(ssh_session session, const char *service)
    int ssh_set_agent_channel(ssh_session session, ssh_channel channel)
    int ssh_set_agent_socket(ssh_session session, socket_t fd)
    void ssh_set_blocking(ssh_session session, int blocking)
    void ssh_set_counters(ssh_session session, ssh_counter scounter, ssh_counter rcounter)
    void ssh_set_fd_except(ssh_session session)
    void ssh_set_fd_toread(ssh_session session)
    void ssh_set_fd_towrite(ssh_session session)
    void ssh_silent_disconnect(ssh_session session)
    int ssh_set_pcap_file(ssh_session session, ssh_pcap_file pcapfile)

    # Userauth
    int ssh_userauth_none(ssh_session session, const char *username)
    int ssh_userauth_list(ssh_session session, const char *username)
    int ssh_userauth_try_publickey(ssh_session session, const char *username, const ssh_key pubkey)
    int ssh_userauth_publickey(ssh_session session, const char *username, const ssh_key privkey)
    # #ifndef _WIN32
    int ssh_userauth_agent(ssh_session session, const char *username)
    # #endif
    int ssh_userauth_publickey_auto(ssh_session session, const char *username, const char *passphrase)
    int ssh_userauth_password(ssh_session session, const char *username, const char *password)

    int ssh_userauth_kbdint(ssh_session session, const char *user, const char *submethods)
    const char *ssh_userauth_kbdint_getinstruction(ssh_session session)
    const char *ssh_userauth_kbdint_getname(ssh_session session)
    int ssh_userauth_kbdint_getnprompts(ssh_session session)
    const char *ssh_userauth_kbdint_getprompt(ssh_session session, unsigned int i, char *echo)
    int ssh_userauth_kbdint_getnanswers(ssh_session session)
    const char *ssh_userauth_kbdint_getanswer(ssh_session session, unsigned int i)
    int ssh_userauth_kbdint_setanswer(ssh_session session, unsigned int i, const char *answer)
    int ssh_userauth_gssapi(ssh_session session)
    const char *ssh_version(int req_version)
    int ssh_write_knownhost(ssh_session session)
    char *ssh_dump_knownhost(ssh_session session)

    void ssh_string_burn(ssh_string str)
    ssh_string ssh_string_copy(ssh_string str)
    void *ssh_string_data(ssh_string str)
    int ssh_string_fill(ssh_string str, const void *data, size_t len)
    void ssh_string_free(ssh_string str)
    ssh_string ssh_string_from_char(const char *what)
    size_t ssh_string_len(ssh_string str)
    ssh_string ssh_string_new(size_t size)
    const char *ssh_string_get_char(ssh_string str)
    char *ssh_string_to_char(ssh_string str)
    void ssh_string_free_char(char *s)

    int ssh_getpass(const char *prompt, char *buf, size_t len, int echo, int verify)

    ctypedef int(*ssh_event_callback)(socket_t fd, int revents, void *userdata)

    ssh_event ssh_event_new()
    int ssh_event_add_fd(ssh_event event, socket_t fd, short events, ssh_event_callback cb, void *userdata)
    int ssh_event_add_session(ssh_event event, ssh_session session)
    int ssh_event_add_connector(ssh_event event, ssh_connector connector)
    int ssh_event_dopoll(ssh_event event, int timeout)
    int ssh_event_remove_fd(ssh_event event, socket_t fd)
    int ssh_event_remove_session(ssh_event event, ssh_session session)
    int ssh_event_remove_connector(ssh_event event, ssh_connector connector)
    void ssh_event_free(ssh_event event)
    const char* ssh_get_clientbanner(ssh_session session)
    const char* ssh_get_serverbanner(ssh_session session)
    const char* ssh_get_kex_algo(ssh_session session)
    const char* ssh_get_cipher_in(ssh_session session)
    const char* ssh_get_cipher_out(ssh_session session)
    const char* ssh_get_hmac_in(ssh_session session)
    const char* ssh_get_hmac_out(ssh_session session)
    ssh_buffer ssh_buffer_new()
    void ssh_buffer_free(ssh_buffer buffer)
    int ssh_buffer_reinit(ssh_buffer buffer)
    int ssh_buffer_add_data(ssh_buffer buffer, const void *data, uint32_t len)
    uint32_t ssh_buffer_get_data(ssh_buffer buffer, void *data, uint32_t requestedlen)
    void *ssh_buffer_get(ssh_buffer buffer)
    uint32_t ssh_buffer_get_len(ssh_buffer buffer)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free
from libc.string cimport memset

from .session cimport Session
from .utils cimport handle_error_codes

from .c_ssh cimport ssh_auth_callback
from . cimport c_callbacks


cdef int auth_callback(const char *prompt, char *buf, size_t len, int echo, int verify, void *userdata):
    try:
        func = <object>userdata
        return func()
    except Exception:
        # TODO - pass back exception
        return -1
    # ssh_getpass(prompt, buf, len, echo, verify);


cdef class Callbacks:

    def __cinit__(self):
        self._cb = <c_callbacks.ssh_callbacks>malloc(
            sizeof(c_callbacks.ssh_callbacks_struct))
        if self._cb is NULL:
            raise MemoryError
        memset(self._cb, 0, sizeof(c_callbacks.ssh_callbacks_struct))
        c_callbacks.ssh_callbacks_init(self._cb)
        # self._cb.userdata = NULL
        self._cb.auth_function = <c_callbacks.ssh_auth_callback>&auth_callback
        # self._cb.log_function = NULL
        # self._cb.connect_status_function = NULL
        # self._cb.global_request_function = NULL
        # self._cb.channel_open_request_x11_function = NULL
        # self._cb.channel_open_request_auth_agent_function = NULL

    def __dealloc__(self):
        if self._cb is not NULL:
            free(self._cb)
            self._cb = NULL

    def set_userdata(self, func not None):
        self._cb.userdata = <void *>func

    def set_callbacks(self, Session session not None):
        cdef int rc
        with nogil:
            rc = c_callbacks.ssh_set_callbacks(session._session, self._cb)
        return handle_error_codes(rc, session._session)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .c_ssh cimport ssh_session, ssh_string, ssh_message, uint32_t, ssh_keytypes_e
from .c_keys cimport ssh_private_key_struct, ssh_public_key_struct

cdef extern from "libssh/legacy.h" nogil:
    ctypedef ssh_private_key_struct* ssh_private_key
    ctypedef ssh_public_key_struct* ssh_public_key
    int ssh_auth_list(ssh_session session)
    int ssh_userauth_offer_pubkey(ssh_session session, const char *username, int type, ssh_string publickey)
    int ssh_userauth_pubkey(ssh_session session, const char *username, ssh_string publickey, ssh_private_key privatekey)
    # IF not _WIN32:
    int ssh_userauth_agent_pubkey(ssh_session session, const char *username, ssh_public_key publickey)
    int ssh_userauth_autopubkey(ssh_session session, const char *passphrase)
    int ssh_userauth_privatekey_file(ssh_session session, const char *username, const char *filename, const char *passphrase)
    void privatekey_free(ssh_private_key prv)
    ssh_private_key privatekey_from_file(ssh_session session, const char *filename, int type, const char *passphrase)
    void publickey_free(ssh_public_key key)
    int ssh_publickey_to_file(ssh_session session, const char *file, ssh_string pubkey, int type)
    ssh_string publickey_from_file(ssh_session session, const char *filename, int *type)
    ssh_public_key publickey_from_privatekey(ssh_private_key prv)
    ssh_string publickey_to_string(ssh_public_key key)
    int ssh_try_publickey_from_file(ssh_session session, const char *keyfile, ssh_string *publickey, int *type)
    ssh_keytypes_e ssh_privatekey_type(ssh_private_key privatekey)
    ssh_string ssh_get_pubkey(ssh_session session)
    ssh_message ssh_message_retrieve(ssh_session session, uint32_t packettype)
    ssh_public_key ssh_message_auth_publickey(ssh_message msg)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free
from libc.string cimport const_char

from .session cimport Session
from .utils cimport to_bytes, to_str, handle_error_codes


from . cimport c_ssh


cdef class Channel:

    def __cinit__(self, Session session):
        self.closed = False
        self._session = session
        self._block_lock = self._session._block_lock

    def __dealloc__(self):
        if self._channel is not NULL and self._session is not None:
            c_ssh.ssh_channel_free(self._channel)
            self._channel = NULL

    @property
    def session(self):
        """Originating session."""
        return self._session

    @staticmethod
    cdef Channel from_ptr(c_ssh.ssh_channel _chan, Session session):
        cdef Channel chan = Channel.__new__(Channel, session)
        chan._channel = _chan
        return chan

    def close(self):
        cdef int rc
        if self.closed:
            return 0
        with nogil:
            rc = c_ssh.ssh_channel_close(self._channel)
        if rc == 0:
            self.closed = True
        return handle_error_codes(rc, self._session._session)

    def get_exit_status(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_get_exit_status(self._channel)
        return rc

    def get_session(self):
        return self.session

    def is_closed(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_channel_is_closed(self._channel)
        return rc != 0

    def is_eof(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_channel_is_eof(self._channel)
        return bool(rc)

    def is_open(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_channel_is_open(self._channel)
        return bool(rc)

    def send_eof(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_send_eof(self._channel)
        return handle_error_codes(rc, self._session._session)

    def request_auth_agent(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_auth_agent(self._channel)
        return handle_error_codes(rc, self._session._session)

    def open_auth_agent(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_open_auth_agent(self._channel)
        return handle_error_codes(rc, self._session._session)

    def open_forward(self, remotehost, int remoteport,
                     sourcehost, int sourceport):
        cdef bytes b_remotehost = to_bytes(remotehost)
        cdef const_char *c_remotehost = b_remotehost
        cdef bytes b_sourcehost = to_bytes(sourcehost)
        cdef const_char *c_sourcehost = b_sourcehost
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_open_forward(self._channel, c_remotehost, remoteport, c_sourcehost, sourceport)
        return handle_error_codes(rc, self._session._session)

    def open_session(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_open_session(self._channel)
        return handle_error_codes(rc, self._session._session)

    def open_x11(self, sourcehost, int sourceport):
        cdef bytes b_sourcehost = to_bytes(sourcehost)
        cdef const_char *c_sourcehost = b_sourcehost
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_open_x11(self._channel, c_sourcehost, sourceport)
        return handle_error_codes(rc, self._session._session)

    def accept_x11(self, int timeout_ms):
        cdef Channel chan
        cdef c_ssh.ssh_channel _chan = NULL
        with nogil:
            _chan = c_ssh.ssh_channel_accept_x11(self._channel, timeout_ms)
        if _chan is NULL:
            raise MemoryError
        chan = Channel.from_ptr(_chan, self._session)
        return chan

    def poll(self, bint is_stderr=False):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_poll(self._channel, is_stderr)
        return handle_error_codes(rc, self._session._session)

    def poll_timeout(self, int timeout, bint is_stderr=False):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_poll_timeout(self._channel, timeout, is_stderr)
        return handle_error_codes(rc, self._session._session)

    def read(self, c_ssh.uint32_t size=1024*1024, bint is_stderr=False):
        cdef int rc
        cdef bytes buf = b''
        cdef char* cbuf
        with nogil:
            cbuf = <char *>malloc(sizeof(char)*size)
            if cbuf is NULL:
                with gil:
                    raise MemoryError
            rc = c_ssh.ssh_channel_read(self._channel, cbuf, size, is_stderr)
        try:
            if rc > 0:
                buf = cbuf[:rc]
        finally:
            free(cbuf)
        return handle_error_codes(rc, self._session._session), buf

    def read_nonblocking(self, c_ssh.uint32_t size=1024*1024, bint is_stderr=False):
        cdef int rc
        cdef bytes buf = b''
        cdef char* cbuf
        with nogil:
            cbuf = <char *>malloc(sizeof(char)*size)
            if cbuf is NULL:
                with gil:
                    raise MemoryError
            rc = c_ssh.ssh_channel_read_nonblocking(self._channel, cbuf, size, is_stderr)
        try:
            if rc > 0:
                buf = cbuf[:rc]
        finally:
            free(cbuf)
        return handle_error_codes(rc, self._session._session), buf

    def read_timeout(self, int timeout, c_ssh.uint32_t size=1024*1024, bint is_stderr=False):
        cdef int rc
        cdef bytes buf = b''
        cdef char* cbuf
        with nogil:
            cbuf = <char *>malloc(sizeof(char)*size)
            if cbuf is NULL:
                with gil:
                    raise MemoryError
            rc = c_ssh.ssh_channel_read_timeout(self._channel, cbuf, size, is_stderr, timeout)
        try:
            if rc > 0:
                buf = cbuf[:rc]
        finally:
            free(cbuf)
        return handle_error_codes(rc, self._session._session), buf

    def request_env(self, name, value):
        cdef bytes b_name = to_bytes(name)
        cdef const_char *c_name = b_name
        cdef bytes b_value = to_bytes(value)
        cdef const_char *c_value = b_value
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_env(self._channel, c_name, c_value)
        return handle_error_codes(rc, self._session._session)

    def request_exec(self, cmd):
        cdef bytes b_cmd = to_bytes(cmd)
        cdef const_char *c_cmd = b_cmd
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_exec(self._channel, c_cmd)
        return handle_error_codes(rc, self._session._session)

    def request_pty(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_pty(self._channel)
        return handle_error_codes(rc, self._session._session)

    def change_pty_size(self, int cols, int rows):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_change_pty_size(self._channel, cols, rows)
        return handle_error_codes(rc, self._session._session)

    def request_pty_size(self, terminal, int col, int row):
        cdef bytes b_terminal = to_bytes(terminal)
        cdef const_char *c_terminal = b_terminal
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_pty_size(self._channel, c_terminal, col, row)
        return handle_error_codes(rc, self._session._session)

    def request_send_signal(self, sig):
        cdef bytes b_sig = to_bytes(sig)
        cdef const_char *c_sig = b_sig
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_send_signal(self._channel, c_sig)
        return handle_error_codes(rc, self._session._session)

    def request_send_break(self, c_ssh.uint32_t length):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_send_break(self._channel, length)
        return handle_error_codes(rc, self._session._session)

    def request_shell(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_shell(self._channel)
        return handle_error_codes(rc, self._session._session)

    def request_sftp(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_sftp(self._channel)
        return handle_error_codes(rc, self._session._session)

    def request_subsystem(self, subsystem):
        cdef bytes b_sys = to_bytes(subsystem)
        cdef const_char *c_sys = b_sys
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_subsystem(self._channel, c_sys)
        return handle_error_codes(rc, self._session._session)

    def request_x11(self, int screen_number, bint single_connection=True):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_channel_request_x11(self._channel, single_connection, NULL, NULL, screen_number)
        return handle_error_codes(rc, self._session._session)

    def set_blocking(self, bint blocking):
        with nogil:
            c_ssh.ssh_channel_set_blocking(self._channel, blocking)

    def set_counter(self, counter):
        raise NotImplementedError

    def write(self, data not None):
        """Write data to stdin on channel.

        :param data: Data to write.
        :type data: str or bytes
        :returns: Return code and bytes written tuples.
        :rtype: (int, int)
        """
        cdef bytes b_buf = to_bytes(data)
        cdef const char *_buf = b_buf
        cdef c_ssh.uint32_t buf_remainder = len(b_buf)
        cdef c_ssh.uint32_t buf_tot_size = buf_remainder
        cdef c_ssh.uint32_t bytes_written = 0
        cdef int rc
        with nogil:
            while buf_remainder > 0:
                rc = c_ssh.ssh_channel_write(self._channel, _buf, buf_remainder)
                if rc < 0 and rc != c_ssh.SSH_AGAIN:
                    with gil:
                        return handle_error_codes(rc, self._session._session)
                elif rc == c_ssh.SSH_AGAIN:
                    break
                _buf += rc
                buf_remainder -= rc
            bytes_written = buf_tot_size - buf_remainder
        return rc, bytes_written

    def write_stderr(self, data not None):
        """Write data to stderr.

        :param data: Data to write.
        :type data: str or bytes
        :returns: Return code and bytes written tuples.
        :rtype: (int, int)
        """
        cdef bytes b_buf = to_bytes(data)
        cdef const char *_buf = b_buf
        cdef c_ssh.uint32_t buf_remainder = len(b_buf)
        cdef c_ssh.uint32_t buf_tot_size = buf_remainder
        cdef c_ssh.uint32_t bytes_written = 0
        cdef int rc
        with nogil:
            while buf_remainder > 0:
                rc = c_ssh.ssh_channel_write_stderr(self._channel, _buf, buf_remainder)
                if rc < 0 and rc != c_ssh.SSH_AGAIN:
                    with gil:
                        return handle_error_codes(rc, self._session._session)
                elif rc == c_ssh.SSH_AGAIN:
                    break
                _buf += rc
                buf_remainder -= rc
            bytes_written = buf_tot_size - buf_remainder
        return rc, bytes_written

    def window_size(self):
        cdef c_ssh.uint32_t size
        with nogil:
            size = c_ssh.ssh_channel_window_size(self._channel)
        return size

    def select(self, channels not None, outchannels not None, maxfd, readfds, timeout=None):
        raise NotImplementedError



# This file is part of RedLibSSH
# Copyright (C) 2017 Panos Kittenis
# Copyright (C) 2022 Red-M

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import time
import threading
import select as pyselect
from . import error_codes
from cpython cimport PyObject_AsFileDescriptor
from libc.stdlib cimport malloc, free
from libc.time cimport time_t
from cython.operator cimport dereference as c_dereference
from libc.string cimport strlen, strdup

from . cimport c_ssh
from . cimport utils

cdef class Tunnel:

    """Tunnel class to provide minor wrapped functions"""

    def __cinit__(self, _session, _tun_channel, sock):
        self._session = _session
        self._tun_channel = _tun_channel
        self._sock = PyObject_AsFileDescriptor(sock)
        self.sock = sock
        self._default_waitsockets = []
        self._waitsockets = []
        self._build_waitsocket_data()
        if self._session.check_c_poll_enabled()==True:
            self._build_c_waitsocket_data()

    # def __dealloc__(self):
        # if self._tun_channel is not NULL and self._session is not None:
            # c_ssh.ssh_channel_free(self._tun_channel)
        # self._tun_channel = NULL

    cdef void _build_c_waitsocket_data(Tunnel self) nogil:
        self._c_waitsockets[0].fd = self._session._sock
        self._c_waitsockets[0].events = 0
        self._c_waitsockets[0].revents = 0
        self._c_waitsockets[1].fd = self._sock
        self._c_waitsockets[1].events = utils.POLLIN | utils.POLLOUT
        self._c_waitsockets[1].revents = 0

    def _build_waitsocket_data(self):
        self._waitsockets = [self._session.sock,self.sock]

    IF HAVE_POLL==1:
        cdef int poll_sockets(Tunnel self,int block_dir,int timeout) nogil:
            cdef int rc

            with nogil:
                if(block_dir & c_ssh.SSH_READ_PENDING):
                    self._c_waitsockets[0].events |= utils.POLLIN

                if(block_dir & c_ssh.SSH_WRITE_PENDING):
                    self._c_waitsockets[0].events |= utils.POLLOUT

                rc = utils.poll(self._c_waitsockets, 2, timeout)
                self._c_waitsockets[0].events = 0
            return rc

    def _block_call(self,_select_timeout=None):
        if _select_timeout==None:
            _select_timeout = 0.005
        with self._session._block_lock:
            # self.keepalive_send()
            block_direction = self._session.get_poll_flags()
        if block_direction==0:
            time.sleep(0.1)
            return([[0,0],[],[]])

        if self._session.check_c_poll_enabled()==True:
            with self._session._block_lock:
                self.poll_sockets(block_direction,_select_timeout*1000)
                return([[self._c_waitsockets[0].revents,self._c_waitsockets[1].revents],[],[]])
        else:
            rfds = self._default_waitsockets
            wfds = self._default_waitsockets
            if block_direction & c_ssh.SSH_READ_PENDING:
                rfds = self._waitsockets

            if block_direction & c_ssh.SSH_WRITE_PENDING:
                wfds = self._waitsockets

            return(pyselect.select(rfds,wfds,self._default_waitsockets,_select_timeout))



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .c_ssh cimport uint64_t, uint32_t, uint8_t, ssh_string, ssh_buffer, ssh_channel, ssh_session, timeval

cdef extern from "fcntl.h" nogil:
    enum:
        O_RDONLY
        O_WRONLY
        O_RDWR
        O_CREAT
        O_EXCL
        O_NOCTTY
        O_NOFOLLOW
        O_TRUNC


cdef extern from "libssh/sftp.h" nogil:
    ctypedef long mode_t
    ctypedef long uid_t
    ctypedef long gid_t
    ctypedef sftp_attributes_struct* sftp_attributes
    ctypedef sftp_client_message_struct* sftp_client_message
    ctypedef sftp_dir_struct* sftp_dir
    # ctypedef sftp_ext_struct *sftp_ext
    ctypedef sftp_file_struct* sftp_file
    ctypedef sftp_message_struct* sftp_message
    ctypedef sftp_packet_struct* sftp_packet
    ctypedef sftp_request_queue_struct* sftp_request_queue
    ctypedef sftp_session_struct* sftp_session
    ctypedef sftp_status_message_struct* sftp_status_message
    ctypedef sftp_statvfs_struct* sftp_statvfs_t

    struct sftp_session_struct:
        ssh_session session
        ssh_channel channel
        int server_version
        int client_version
        int version
        sftp_request_queue queue
        uint32_t id_counter
        int errnum
        void **handles
        # sftp_ext ext

    struct sftp_packet_struct:
        sftp_session sftp
        uint8_t type
        ssh_buffer payload

    struct sftp_file_struct:
        sftp_session sftp
        char *name
        uint64_t offset
        ssh_string handle
        int eof
        int nonblocking

    struct sftp_dir_struct:
        sftp_session sftp
        char *name
        ssh_string handle
        ssh_buffer buffer
        uint32_t count
        int eof

    struct sftp_message_struct:
        sftp_session sftp
        uint8_t packet_type
        ssh_buffer payload
        uint32_t id

    struct sftp_client_message_struct:
        sftp_session sftp
        uint8_t type
        uint32_t id
        char *filename
        uint32_t flags
        sftp_attributes attr
        ssh_string handle
        uint64_t offset
        uint32_t len
        int attr_num
        ssh_buffer attrbuf
        ssh_string data
        ssh_buffer complete_message
        char *str_data

    struct sftp_request_queue_struct:
        sftp_request_queue _next "next"
        sftp_message message

    struct sftp_status_message_struct:
        uint32_t _id "id"
        uint32_t status
        ssh_string error_unused
        ssh_string lang_unused
        char *errormsg
        char *langmsg

    struct sftp_attributes_struct:
        char *name
        char *longname
        uint32_t flags
        uint8_t type
        uint64_t size
        uint32_t uid
        uint32_t gid
        char *owner
        char *group
        uint32_t permissions
        uint64_t atime64
        uint32_t atime
        uint32_t atime_nseconds
        uint64_t createtime
        uint32_t createtime_nseconds
        uint64_t mtime64
        uint32_t mtime
        uint32_t mtime_nseconds
        ssh_string acl
        uint32_t extended_count
        ssh_string extended_type
        ssh_string extended_data

    struct sftp_statvfs_struct:
        uint64_t f_bsize
        uint64_t f_frsize
        uint64_t f_blocks
        uint64_t f_bfree
        uint64_t f_bavail
        uint64_t f_files
        uint64_t f_ffree
        uint64_t f_favail
        uint64_t f_fsid
        uint64_t f_flag
        uint64_t f_namemax
    sftp_session sftp_new(ssh_session session)
    sftp_session sftp_new_channel(ssh_session session, ssh_channel channel)
    void sftp_free(sftp_session sftp)
    int sftp_init(sftp_session sftp)
    int sftp_get_error(sftp_session sftp)
    unsigned int sftp_extensions_get_count(sftp_session sftp)
    const char *sftp_extensions_get_name(sftp_session sftp, unsigned int indexn)
    const char *sftp_extensions_get_data(sftp_session sftp, unsigned int indexn)
    int sftp_extension_supported(sftp_session sftp, const char *name, const char *data)
    sftp_dir sftp_opendir(sftp_session session, const char *path)
    sftp_attributes sftp_readdir(sftp_session session, sftp_dir dir)
    int sftp_dir_eof(sftp_dir dir)
    sftp_attributes sftp_stat(sftp_session session, const char *path)
    sftp_attributes sftp_lstat(sftp_session session, const char *path)
    sftp_attributes sftp_fstat(sftp_file file)
    void sftp_attributes_free(sftp_attributes file)
    int sftp_closedir(sftp_dir dir)
    int sftp_close(sftp_file file)
    sftp_file sftp_open(sftp_session session, const char *file, int accesstype, mode_t mode)
    void sftp_file_set_nonblocking(sftp_file handle)
    void sftp_file_set_blocking(sftp_file handle)
    ssize_t sftp_read(sftp_file file, void *buf, size_t count)
    int sftp_async_read_begin(sftp_file file, uint32_t len)
    int sftp_async_read(sftp_file file, void *data, uint32_t len, uint32_t id)
    ssize_t sftp_write(sftp_file file, const void *buf, size_t count)
    int sftp_seek(sftp_file file, uint32_t new_offset)
    int sftp_seek64(sftp_file file, uint64_t new_offset)
    unsigned long sftp_tell(sftp_file file)
    uint64_t sftp_tell64(sftp_file file)
    void sftp_rewind(sftp_file file)
    int sftp_unlink(sftp_session sftp, const char *file)
    int sftp_rmdir(sftp_session sftp, const char *directory)
    int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
    int sftp_rename(sftp_session sftp, const char *original, const  char *newname)
    int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr)
    int sftp_chown(sftp_session sftp, const char *file, uid_t owner, gid_t group)
    int sftp_chmod(sftp_session sftp, const char *file, mode_t mode)
    int sftp_utimes(sftp_session sftp, const char *file, const timeval *times)
    int sftp_symlink(sftp_session sftp, const char *target, const char *dest)
    char *sftp_readlink(sftp_session sftp, const char *path)
    sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path)
    sftp_statvfs_t sftp_fstatvfs(sftp_file file)
    void sftp_statvfs_free(sftp_statvfs_t statvfs_o)
    int sftp_fsync(sftp_file file)
    char *sftp_canonicalize_path(sftp_session sftp, const char *path)
    int sftp_server_version(sftp_session sftp)

    # # Server
    # sftp_session sftp_server_new(ssh_session session, ssh_channel chan)
    # int sftp_server_init(sftp_session sftp)
    enum:
        SSH_FXP_INIT
        SSH_FXP_VERSION
        SSH_FXP_OPEN
        SSH_FXP_CLOSE
        SSH_FXP_READ
        SSH_FXP_WRITE
        SSH_FXP_LSTAT
        SSH_FXP_FSTAT
        SSH_FXP_SETSTAT
        SSH_FXP_FSETSTAT
        SSH_FXP_OPENDIR
        SSH_FXP_READDIR
        SSH_FXP_REMOVE
        SSH_FXP_MKDIR
        SSH_FXP_RMDIR
        SSH_FXP_REALPATH
        SSH_FXP_STAT
        SSH_FXP_RENAME
        SSH_FXP_READLINK
        SSH_FXP_SYMLINK
        SSH_FXP_STATUS
        SSH_FXP_HANDLE
        SSH_FXP_DATA
        SSH_FXP_NAME
        SSH_FXP_ATTRS
        SSH_FXP_EXTENDED
        SSH_FXP_EXTENDED_REPLY
    enum:
        SSH_FILEXFER_ATTR_SIZE
        SSH_FILEXFER_ATTR_PERMISSIONS
        SSH_FILEXFER_ATTR_ACCESSTIME
        SSH_FILEXFER_ATTR_ACMODTIME
        SSH_FILEXFER_ATTR_CREATETIME
        SSH_FILEXFER_ATTR_MODIFYTIME
        SSH_FILEXFER_ATTR_ACL
        SSH_FILEXFER_ATTR_OWNERGROUP
        SSH_FILEXFER_ATTR_SUBSECOND_TIMES
        SSH_FILEXFER_ATTR_EXTENDED
        SSH_FILEXFER_ATTR_UIDGID
    # Types
    enum:
        SSH_FILEXFER_TYPE_REGULAR
        SSH_FILEXFER_TYPE_DIRECTORY
        SSH_FILEXFER_TYPE_SYMLINK
        SSH_FILEXFER_TYPE_SPECIAL
        SSH_FILEXFER_TYPE_UNKNOWN
    enum:
        SSH_FX_OK
        SSH_FX_EOF
        SSH_FX_NO_SUCH_FILE
        SSH_FX_PERMISSION_DENIED
        SSH_FX_FAILURE
        SSH_FX_BAD_MESSAGE
        SSH_FX_NO_CONNECTION
        SSH_FX_CONNECTION_LOST
        SSH_FX_OP_UNSUPPORTED
        SSH_FX_INVALID_HANDLE
        SSH_FX_NO_SUCH_PATH
        SSH_FX_FILE_ALREADY_EXISTS
        SSH_FX_WRITE_PROTECT
        SSH_FX_NO_MEDIA
    enum:
        SSH_FXF_READ
        SSH_FXF_WRITE
        SSH_FXF_APPEND
        SSH_FXF_CREAT
        SSH_FXF_TRUNC
        SSH_FXF_EXCL
        SSH_FXF_TEXT
    enum:
        SSH_S_IFMT
        SSH_S_IFSOCK
        SSH_S_IFLNK
        SSH_S_IFREG
        SSH_S_IFBLK
        SSH_S_IFDIR
        SSH_S_IFCHR
        SSH_S_IFIFO
    enum:
        SSH_FXF_RENAME_OVERWRITE
        SSH_FXF_RENAME_ATOMIC
        SSH_FXF_RENAME_NATIVE
    enum:
        SFTP_OPEN
        SFTP_CLOSE
        SFTP_READ
        SFTP_WRITE
        SFTP_LSTAT
        SFTP_FSTAT
        SFTP_SETSTAT
        SFTP_FSETSTAT
        SFTP_OPENDIR
        SFTP_READDIR
        SFTP_REMOVE
        SFTP_MKDIR
        SFTP_RMDIR
        SFTP_REALPATH
        SFTP_STAT
        SFTP_RENAME
        SFTP_READLINK
        SFTP_SYMLINK
    enum:
        SSH_FXE_STATVFS_ST_RDONLY
        SSH_FXE_STATVFS_ST_NOSUID



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .sftp cimport SFTP

from . cimport c_sftp


cdef class SFTPStatVFS:

    def __dealloc__(self):
        if self._stats is not NULL:
            c_sftp.sftp_statvfs_free(self._stats)
            self._stats = NULL

    @staticmethod
    cdef SFTPStatVFS from_ptr(c_sftp.sftp_statvfs_t stats, SFTP sftp):
        cdef SFTPStatVFS _vfs = SFTPStatVFS.__new__(SFTPStatVFS)
        _vfs._stats = stats
        _vfs.sftp = sftp
        return _vfs

    @property
    def f_bsize(self):
        return self._stats.f_bsize if self._stats is not NULL else None

    @property
    def f_frsize(self):
        return self._stats.f_frsize if self._stats is not NULL else None

    @property
    def f_blocks(self):
        return self._stats.f_blocks if self._stats is not NULL else None

    @property
    def f_bfree(self):
        return self._stats.f_bfree if self._stats is not NULL else None

    @property
    def f_bavail(self):
        return self._stats.f_bavail if self._stats is not NULL else None

    @property
    def f_files(self):
        return self._stats.f_files if self._stats is not NULL else None

    @property
    def f_ffree(self):
        return self._stats.f_ffree if self._stats is not NULL else None

    @property
    def f_favail(self):
        return self._stats.f_favail if self._stats is not NULL else None

    @property
    def f_fsid(self):
        return self._stats.f_fsid if self._stats is not NULL else None

    @property
    def f_flag(self):
        return self._stats.f_flag if self._stats is not NULL else None

    @property
    def f_namemax(self):
        return self._stats.f_namemax if self._stats is not NULL else None



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free

from .sftp_attributes cimport SFTPAttributes
from .sftp_statvfs cimport SFTPStatVFS

from .exceptions import SFTPError, SFTPHandleError

from .c_ssh cimport uint32_t, uint64_t, ssh_get_error, SSH_ERROR, SSH_AGAIN
from . cimport c_sftp


cdef class SFTPFile:

    @staticmethod
    cdef SFTPFile from_ptr(c_sftp.sftp_file _file, SFTP sftp):
        cdef SFTPFile _fh = SFTPFile.__new__(SFTPFile, sftp)
        _fh._file = _file
        _fh.sftp = sftp
        return _fh

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        size, data = self.read()
        if size > 0:
            return size, data
        raise StopIteration

    def fstat(self):
        cdef SFTPAttributes _attrs
        cdef c_sftp.sftp_attributes c_attrs
        with nogil:
            c_attrs = c_sftp.sftp_fstat(self._file)
        if c_attrs is NULL:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        _attrs = SFTPAttributes.from_ptr(c_attrs, self.sftp)
        return _attrs

    def close(self):
        cdef int rc
        if self.closed:
            return 0
        with nogil:
            rc = c_sftp.sftp_close(self._file)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        self.closed = True
        return rc

    def set_nonblocking(self):
        with nogil:
            c_sftp.sftp_file_set_nonblocking(self._file)

    def set_blocking(self):
        with nogil:
            c_sftp.sftp_file_set_blocking(self._file)

    def read(self, size_t size=1048576):
        cdef ssize_t _size
        cdef bytes buf = b''
        cdef char *c_buf
        with nogil:
            c_buf = <char *>malloc(sizeof(char) * size)
            if c_buf is NULL:
                with gil:
                    raise MemoryError
            _size = c_sftp.sftp_read(self._file, c_buf, size)
        try:
            if _size > 0:
                buf = c_buf[:_size]
        finally:
            free(c_buf)
        return _size, buf

    def async_read_begin(self, uint32_t length=1048576):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_async_read_begin(self._file, length)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc

    def async_read(self, uint32_t _id, uint32_t length=1048576):
        cdef int size
        cdef bytes buf = b''
        cdef char *c_buf
        with nogil:
            c_buf = <char *>malloc(sizeof(char) * length)
            if c_buf is NULL:
                with gil:
                    raise MemoryError
            size = c_sftp.sftp_async_read(self._file, c_buf, length, _id)
        try:
            if size > 0:
                buf = c_buf[:size]
            elif size < 0:
                if size == SSH_ERROR:
                    raise SFTPError(ssh_get_error(self.sftp.session._session))
                elif size == SSH_AGAIN:
                    return SSH_AGAIN, buf
        finally:
            free(c_buf)
        return size, buf

    def write(self, bytes data):
        cdef ssize_t rc
        cdef const char *c_data = data
        cdef size_t data_len = len(data)
        with nogil:
            rc = c_sftp.sftp_write(self._file, c_data, data_len)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.sftp.session._session))
        return rc

    def seek(self, uint32_t offset):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_seek(self._file, offset)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc

    def seek64(self, uint64_t offset):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_seek64(self._file, offset)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc

    def tell(self):
        cdef unsigned long rc
        with nogil:
            rc = c_sftp.sftp_tell(self._file)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc

    def tell64(self):
        cdef uint64_t rc
        with nogil:
            rc = c_sftp.sftp_tell64(self._file)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc

    def rewind(self):
        with nogil:
            c_sftp.sftp_rewind(self._file)

    def fstatvfs(self):
        cdef SFTPStatVFS vfs
        cdef c_sftp.sftp_statvfs_t c_vfs
        with nogil:
            c_vfs = c_sftp.sftp_fstatvfs(self._file)
        if c_vfs is NULL:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        vfs = SFTPStatVFS.from_ptr(c_vfs, self.sftp)
        return vfs

    def fsync(self):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_fsync(self._file)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        return rc


cdef class SFTPDir:

    def __dealloc__(self):
        if self.closed==False and self.sftp._sftp is not NULL and self.sftp.session._session is not NULL:
            self.closedir()

    @staticmethod
    cdef SFTPDir from_ptr(c_sftp.sftp_dir _dir, SFTP sftp):
        cdef SFTPDir _fh = SFTPDir.__new__(SFTPDir, sftp)
        _fh._dir = _dir
        _fh.sftp = sftp
        return _fh

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.closedir()

    def __iter__(self):
        return self

    def __next__(self):
        cdef SFTPAttributes _attrs
        _attrs = self.readdir()
        while _attrs is not None:
            return _attrs
        raise StopIteration

    @property
    def sftp_session(self):
        return self.sftp

    def eof(self):
        cdef bint rc
        with nogil:
            rc = c_sftp.sftp_dir_eof(self._dir)
        return bool(rc)

    def closedir(self):
        cdef int rc
        if self.closed:
            return 0
        with nogil:
            rc = c_sftp.sftp_closedir(self._dir)
        if rc < 0:
            raise SFTPHandleError(ssh_get_error(self.sftp.session._session))
        self.closed = True
        return rc

    cpdef SFTPAttributes readdir(self):
        cdef SFTPAttributes _attrs
        cdef c_sftp.sftp_attributes c_attrs
        with nogil:
            c_attrs = c_sftp.sftp_readdir(self.sftp._sftp, self._dir)
        if c_sftp.sftp_dir_eof(self._dir) == 1:
            return
        elif c_attrs is NULL:
            raise SFTPHandleError
        _attrs = SFTPAttributes.from_ptr(c_attrs, self.sftp)
        return _attrs



import platform
import os
import sys
from glob import glob
from _setup_libssh import build_ssh

import versioneer
from setuptools import setup, find_packages

cpython = platform.python_implementation() == 'CPython'

try:
    from Cython.Distutils.extension import Extension
    from Cython.Distutils import build_ext
except ImportError:
    from setuptools import Extension
    USING_CYTHON = False
else:
    USING_CYTHON = True

_PYTHON_MAJOR_VERSION = int(platform.python_version_tuple()[0])
ON_WINDOWS = platform.system() == 'Windows'
ON_LINUX = platform.system() == 'Linux'
ON_OSX = platform.system() == 'Darwin'
ON_RTD = os.environ.get('READTHEDOCS') == 'True'
REDLIBSSH_BUILD_TRACING = bool(os.environ.get('REDLIBSSH_BUILD_TRACING', 0))
SYSTEM_LIBSSH = bool(os.environ.get('SYSTEM_LIBSSH', 0)) or ON_RTD or ON_WINDOWS

if ON_WINDOWS and _PYTHON_MAJOR_VERSION < 3:
    raise ImportError(
        "ssh-python requires Python 3 or above on Windows platforms.")

# Only build libssh if SYSTEM_LIBSSH is not set and running a build
if not SYSTEM_LIBSSH and ON_LINUX and (len(sys.argv) >= 2 and not (
        '--help' in sys.argv[1:] or
        sys.argv[1] in (
            '--help-commands', 'egg_info', '--version', 'clean',
            'sdist', '--long-description')) and
                          __name__ == '__main__'):
    build_ssh()

ext = 'pyx' if USING_CYTHON else 'c'
sources = glob('ssh/*.%s' % (ext,))
_arch = platform.architecture()[0][0:2]
_libs = ['ssh'] if not ON_WINDOWS else [
    'ssh', 'Ws2_32', 'user32',
    'libcrypto%sMD' % _arch, 'libssl%sMD' % _arch,
    'zlibstatic',
]

# _comp_args = ["-ggdb"]
_comp_args = ["-O3"] if not ON_WINDOWS else None
cython_directives = {
    'embedsignature': True,
    'boundscheck': False,
    'optimize.use_switch': True,
    'wraparound': False,
    'language_level': 3,
}
cython_args = {}
if USING_CYTHON:
    cython_args = {
        'cython_directives': cython_directives,
        'cython_compile_time_env': {
            'ON_WINDOWS': ON_WINDOWS,
            'HAVE_POLL': ON_LINUX,
        }
    }
    if REDLIBSSH_BUILD_TRACING==True:
        cython_args.update({'define_macros':[("CYTHON_TRACE", 1),("CYTHON_TRACE_NOGIL", 1)]})
        cython_args['cython_directives'].update({'linetrace':True})


runtime_library_dirs = ["$ORIGIN/."] if not SYSTEM_LIBSSH or ON_OSX else None
_lib_dir = os.path.abspath("./local/lib") if not SYSTEM_LIBSSH or ON_OSX else "/usr/local/lib"
include_dirs = ["./local/include"] if (ON_WINDOWS or ON_RTD) or not SYSTEM_LIBSSH or ON_OSX else ["/usr/local/include"]

extensions = [
    Extension(
        sources[i].split('.')[0].replace(os.path.sep, '.'),
        sources=[sources[i]],
        include_dirs=include_dirs,
        libraries=_libs,
        library_dirs=[_lib_dir,os.path.abspath("./local/lib")],
        runtime_library_dirs=runtime_library_dirs,
        extra_compile_args=_comp_args,
        **cython_args
    )
    for i in range(len(sources))]

package_data = {'ssh': ['*.pxd', 'libssh.so*']}

if ON_WINDOWS:
    package_data['ssh'].extend([
        'libcrypto*.dll', 'libssl*.dll',
        'ssh.dll', 'msvc*.dll', 'vcruntime*.dll',
    ])

cmdclass = versioneer.get_cmdclass()
if USING_CYTHON:
    cmdclass['build_ext'] = build_ext
    sys.stdout.write("Cython arguments: %s%s" % (cython_args, os.linesep))


sys.stdout.write("Windows platform: %s, Python major version: %s.%s" % (ON_WINDOWS, _PYTHON_MAJOR_VERSION, os.sep))
if ON_WINDOWS and _PYTHON_MAJOR_VERSION == 2:
    # Python 2 on Windows builds are unsupported.
    extensions = [
        Extension(
            'ssh',
            sources=[os.sep.join(['ssh', '__init__.%s' % (ext,)])],
            extra_compile_args=_comp_args,
            **cython_args
        )
    ]
    package_data = {}
    sys.stdout.write("Windows Python 2 build - %s extensions: %s" % (len(extensions), os.sep))

install_requires = []

test_deps = [
    'coveralls',
    'pytest-cov',
    'pylint',
    'bandit',
    'safety'
]

setup(
    name='redlibssh',
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    url='https://github.com/Red-M/Redlibssh',
    license='LGPLv2.1',
    author='Red_M',
    author_email='redlibssh_pypi@red-m.net',
    description=('Alternate bindings for libssh C library'),
    long_description=open('README.rst').read(),
    packages=find_packages(
        '.', exclude=('embedded_server', 'embedded_server.*',
                      'tests', 'tests.*',
                      '*.tests', '*.tests.*')),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=install_requires,
    extras_require={
        'tests':list(set(install_requires+test_deps))
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: System :: Shells',
        'Topic :: System :: Networking',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Operating System :: MacOS :: MacOS X',
    ],
    ext_modules=extensions,
    package_data=package_data,
)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from cpython cimport PyObject_AsFileDescriptor

from .channel cimport Channel
from .session cimport Session

from .utils cimport handle_error_codes
from . cimport c_ssh


cdef class Flag:

    @staticmethod
    cdef Flag from_flag(c_ssh.ssh_connector_flags_e flag):
        cdef Flag _flag = Flag.__new__(Flag)
        _flag._flag = flag
        return _flag

    def __eq__(self, Flag other not None):
        return self._flag == other._flag

    def __str__(self):
        return str(self._flag)

    def __repr__(self):
        return self.__str__()


CONNECTOR_STDOUT = Flag.from_flag(c_ssh.ssh_connector_flags_e.SSH_CONNECTOR_STDOUT)
CONNECTOR_STDERR = Flag.from_flag(c_ssh.ssh_connector_flags_e.SSH_CONNECTOR_STDERR)
CONNECTOR_BOTH = Flag.from_flag(c_ssh.ssh_connector_flags_e.SSH_CONNECTOR_BOTH)


cdef class Connector:

    def __cinit__(self, Session session):
        self.session = session

    def __dealloc__(self):
        if self._connector is not NULL:
            c_ssh.ssh_connector_free(self._connector)
            self._connector = NULL

    @staticmethod
    cdef Connector from_ptr(c_ssh.ssh_connector _connector, Session session):
        cdef Connector connector = Connector.__new__(Connector, session)
        connector._connector = _connector
        return connector

    def set_in_channel(self, Channel channel, Flag flag):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_connector_set_in_channel(self._connector, channel._channel, flag._flag)
        return handle_error_codes(rc, self.session._session)

    def set_out_channel(self, Channel channel, Flag flag):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_connector_set_out_channel(self._connector, channel._channel, flag._flag)
        return handle_error_codes(rc, self.session._session)

    def set_in_fd(self, socket):
        cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(socket)
        with nogil:
            c_ssh.ssh_connector_set_in_fd(self._connector, _sock)

    def set_out_fd(self, socket):
        cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(socket)
        with nogil:
            c_ssh.ssh_connector_set_out_fd(self._connector, _sock)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .c_ssh cimport ssh_options_e


cdef class Option:
    cdef ssh_options_e _option

    @staticmethod
    cdef Option from_option(ssh_options_e option)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from . cimport c_ssh
IF HAVE_POLL==1:
    from . cimport utils

cdef class Session:
    cdef c_ssh.ssh_session _session
    cdef c_ssh.socket_t _sock
    cdef readonly object sock

    cdef readonly bint c_poll_enabled
    cdef readonly object _default_waitsockets
    cdef readonly object _waitsockets
    cdef public bint c_poll_use
    cdef public object _block_lock

    cdef void _build_c_waitsocket_data(Session self) nogil
    IF HAVE_POLL==1:
        cdef int poll_socket(Session self,int block_dir,int timeout) nogil
        cdef readonly utils.pollfd _c_waitsockets[1]



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free

from .session cimport Session
from .sftp_handles cimport SFTPFile, SFTPDir
from .sftp_attributes cimport SFTPAttributes
from .sftp_statvfs cimport SFTPStatVFS
from .utils cimport handle_error_codes, to_bytes, to_str
from .exceptions import SFTPError, SFTPHandleError

from .c_ssh cimport ssh_get_error, ssh_get_error_code, timeval
from . cimport c_sftp


cdef class SFTP:

    @staticmethod
    cdef SFTP from_ptr(c_sftp.sftp_session sftp, Session session):
        cdef SFTP _sftp = SFTP.__new__(SFTP)
        _sftp._sftp = sftp
        _sftp.session = session
        return _sftp

    def __dealloc__(self):
        if self._sftp is not NULL:
            c_sftp.sftp_free(self._sftp)
            self._sftp = NULL

    def init(self):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_init(self._sftp)
        return handle_error_codes(rc, self.session._session)

    def get_error(self):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_get_error(self._sftp)
        return rc

    def extensions_get_count(self):
        cdef unsigned int rc
        with nogil:
            rc = c_sftp.sftp_extensions_get_count(self._sftp)
        return rc

    def extensions_get_name(self, unsigned int indexn):
        cdef const char *_name
        cdef bytes name = None
        with nogil:
            _name = c_sftp.sftp_extensions_get_name(self._sftp, indexn)
        if _name is not NULL:
            name = _name
        return name

    def extensions_get_data(self, unsigned int indexn):
        cdef const char *_name
        cdef bytes name = None
        with nogil:
            _name = c_sftp.sftp_extensions_get_data(self._sftp, indexn)
        if _name is not NULL:
            name = _name
        return name

    def extension_supported(self, name not None, data not None):
        cdef bint rc
        cdef bytes b_name = to_bytes(name)
        cdef const char *c_name = b_name
        cdef bytes b_data = to_bytes(data)
        cdef const char *c_data = b_data
        with nogil:
            rc = c_sftp.sftp_extension_supported(self._sftp, c_name, c_data)
        return bool(rc)

    def opendir(self, path not None):
        cdef SFTPDir _dir
        cdef c_sftp.sftp_dir c_dir
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        with nogil:
            c_dir = c_sftp.sftp_opendir(self._sftp, c_path)
        if c_dir is NULL:
            raise SFTPHandleError(ssh_get_error(self.session._session))
        _dir = SFTPDir.from_ptr(c_dir, self)
        return _dir

    def stat(self, path not None):
        cdef SFTPAttributes _attrs
        cdef c_sftp.sftp_attributes c_attrs
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        with nogil:
            c_attrs = c_sftp.sftp_stat(self._sftp, c_path)
        if c_attrs is NULL:
            raise SFTPError(ssh_get_error(self.session._session))
        _attrs = SFTPAttributes.from_ptr(c_attrs, self)
        return _attrs

    def lstat(self, path not None):
        cdef SFTPAttributes _attrs
        cdef c_sftp.sftp_attributes c_attrs
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        with nogil:
            c_attrs = c_sftp.sftp_lstat(self._sftp, c_path)
        if c_attrs is NULL:
            raise SFTPError(ssh_get_error(self.session._session))
        _attrs = SFTPAttributes.from_ptr(c_attrs, self)
        return _attrs

    def open(self, path not None, int accesstype, c_sftp.mode_t mode):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef SFTPFile _file
        cdef c_sftp.sftp_file c_file
        with nogil:
            c_file = c_sftp.sftp_open(self._sftp, c_path, accesstype, mode)
        if c_file is NULL:
            raise SFTPHandleError(ssh_get_error(self.session._session))
        _file = SFTPFile.from_ptr(c_file, self)
        return _file

    def unlink(self, path not None):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_unlink(self._sftp, c_path)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def rmdir(self, path not None):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_rmdir(self._sftp, c_path)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def mkdir(self, path not None, c_sftp.mode_t mode):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_mkdir(self._sftp, c_path, mode)
        return handle_error_codes(rc, self.session._session)

    def rename(self, original not None, newname not None):
        cdef bytes b_orig = to_bytes(original)
        cdef const char *c_orig = b_orig
        cdef bytes b_newname = to_bytes(newname)
        cdef const char *c_newname = b_newname
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_rename(self._sftp, c_orig, c_newname)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def setstat(self, path not None, SFTPAttributes attr):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_setstat(self._sftp, c_path, attr._attrs)
        return handle_error_codes(rc, self.session._session)

    def chown(self, path not None,
              c_sftp.uid_t owner, c_sftp.gid_t group):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_chown(self._sftp, c_path, owner, group)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def chmod(self, path not None, c_sftp.mode_t mode):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_chmod(self._sftp, c_path, mode)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def utimes(self, path not None, long seconds, long microseconds):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef int rc
        cdef timeval *_val
        with nogil:
            _val = <timeval *>malloc(sizeof(timeval))
            if _val is NULL:
                with gil:
                    raise MemoryError
            _val.tv_sec = seconds
            _val.tv_usec = microseconds
            rc = c_sftp.sftp_utimes(self._sftp, c_path, _val)
            free(_val)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def symlink(self, source not None, dest not None):
        cdef bytes b_source = to_bytes(source)
        cdef const char *c_source = b_source
        cdef bytes b_dest = to_bytes(dest)
        cdef const char *c_dest = b_dest
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_symlink(self._sftp, c_source, c_dest)
        if rc < 0:
            raise SFTPError(ssh_get_error(self.session._session))
        return rc

    def readlink(self, path not None):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef char *_link
        cdef bytes b_link
        with nogil:
            _link = c_sftp.sftp_readlink(self._sftp, c_path)
        if _link is NULL:
            raise SFTPError(ssh_get_error(self.session._session))
        b_link = _link
        return to_str(b_link)

    def statvfs(self, path not None):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef SFTPStatVFS vfs
        cdef c_sftp.sftp_statvfs_t c_vfs
        with nogil:
            c_vfs = c_sftp.sftp_statvfs(self._sftp, c_path)
        if c_vfs is NULL:
            raise SFTPError(ssh_get_error(self.session._session))
        vfs = SFTPStatVFS.from_ptr(c_vfs, self)
        return vfs

    def canonicalize_path(self, path not None):
        cdef bytes b_path = to_bytes(path)
        cdef const char *c_path = b_path
        cdef char *_rpath
        cdef bytes b_rpath
        with nogil:
            _rpath = c_sftp.sftp_canonicalize_path(self._sftp, c_path)
        if _rpath is NULL:
            raise SFTPError(ssh_get_error(self.session._session))
        b_rpath = _rpath
        return to_str(b_rpath)

    def server_version(self):
        cdef int rc
        with nogil:
            rc = c_sftp.sftp_server_version(self._sftp)
        return handle_error_codes(rc, self.session._session)
